class BasicDatabaseRouter:
    route_app_labels = set()
    database_name = None

    def db_for_read(self, model, **hints):
        if model._meta.app_label in self.route_app_labels:
            return self.database_name
        return None

    def db_for_write(self, model, **hints):
        if model._meta.app_label in self.route_app_labels:
            return self.database_name
        return None

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        if app_label in self.route_app_labels:
            return db == self.database_name
        return None


class AudomaApiRouter(BasicDatabaseRouter):
    route_app_labels = {"audoma_api"}
    database_name = "audoma_api"


class HealthCareApiRouter(BasicDatabaseRouter):
    route_app_labels = {"healthcare_api"}
    database_name = "healthcare_api"



"""
This module is an override for default drf's field module.
Most of those fields, are providing additional example functionality,
and also has defined schema type.
"""

import sys
from typing import Any

import exrex
import phonenumbers
from djmoney.contrib.django_rest_framework import MoneyField
from drf_extra_fields import fields as extra_fields
from drf_extra_fields.fields import *  # noqa: F403, F401
from drf_spectacular.drainage import set_override
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import extend_schema_field
from phonenumber_field import serializerfields
from phonenumber_field.phonenumber import to_python
from psycopg2._range import (
    DateRange,
    DateTimeTZRange,
    NumericRange,
)
from rest_framework import fields
from rest_framework.fields import *  # noqa: F403, F401

from django.core import validators

from audoma.example_generators import generate_lorem_ipsum
from audoma.mixins import (
    Base64ExampleMixin,
    DateExampleMixin,
    DateTimeExampleMixin,
    ExampleMixin,
    NumericExampleMixin,
    RangeExampleMixin,
    RegexExampleMixin,
    TimeExampleMixin,
)


field_names = [
    "BooleanField",
    "NullBooleanField",
    "EmailField",
    "SlugField",
    "URLField",
    "DurationField",
    "ChoiceField",
    "MultipleChoiceField",
    "FilePathField",
    "FileField",
    "ImageField",
    "ListField",
    "DictField",
    "HStoreField",
    "JSONField",
    "ReadOnlyField",
]


this = sys.modules[__name__]


for field_name in field_names:
    setattr(
        this,
        field_name,
        type(field_name, (ExampleMixin, getattr(fields, field_name)), {}),
    )


class DecimalField(NumericExampleMixin, fields.DecimalField):
    pass


@extend_schema_field(OpenApiTypes.UUID)
class UUIDField(ExampleMixin, fields.UUIDField):
    pass


class IntegerField(NumericExampleMixin, fields.IntegerField):
    pass


class FloatField(NumericExampleMixin, fields.FloatField):
    pass


class RegexField(RegexExampleMixin, fields.RegexField):
    pass


class MACAddressField(RegexExampleMixin, fields.CharField):
    def __init__(self, **kwargs) -> None:
        self.regex = "^([0-9A-F]{2}:){5}([0-9A-F]{2})|([0-9A-F]{2}-){5}([0-9A-F]{2})$"
        self.validators = [validators.RegexValidator(self.regex)]
        super().__init__(**kwargs)


@extend_schema_field(OpenApiTypes.DATE)
class DateField(DateExampleMixin, fields.DateField):
    pass


@extend_schema_field(OpenApiTypes.TIME)
class TimeField(TimeExampleMixin, fields.TimeField):
    pass


class DateTimeField(DateTimeExampleMixin, fields.DateTimeField):
    ...


@extend_schema_field(
    field={
        "format": "ip-address",
        "example": str(
            exrex.getone(
                r"^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$"
            )
        ),
    }
)
class IPAddressField(ExampleMixin, fields.IPAddressField):
    pass


@extend_schema_field(field={"format": "tel"})
class PhoneNumberField(ExampleMixin, serializerfields.PhoneNumberField):
    def __init__(self, *args, **kwargs) -> None:
        example = kwargs.pop("example", None)
        if example is None:
            number = phonenumbers.example_number(None)
            example = str(to_python(number))
        super().__init__(*args, example=example, **kwargs)


class CharField(ExampleMixin, fields.CharField):
    def __init__(self, *args, **kwargs) -> None:
        example = kwargs.pop("example", None)
        min_length = kwargs.get("min_length", 20)
        max_length = kwargs.get("max_length", 80)
        if not example:
            example = generate_lorem_ipsum(min_length=min_length, max_length=max_length)

        super().__init__(*args, example=example, **kwargs)


class MoneyField(NumericExampleMixin, MoneyField):
    pass


class ChoiceField(ExampleMixin, fields.ChoiceField):
    def __init__(self, choices, **kwargs):
        self.original_choices = fields.to_choices_dict(choices)
        super().__init__(choices, **kwargs)


class SerializerMethodField(ExampleMixin, fields.Field):
    def _parse_field(self, field):
        if field is not None and not isinstance(field, fields.Field):
            raise ValueError(
                f"Incorrect type of field, field \
                    must be an instance of rest_framework.fields.Field.\
                        Passed value: {field}"
            )
        return field

    def __init__(
        self, *args, method_name=None, field=None, writable=False, **kwargs
    ) -> None:
        self.method_name = method_name
        self.field = self._parse_field(field)
        writable = writable
        if writable and self.field is None:
            raise ValueError("Writable SerializerMethodField must have field defined.")

        kwargs["source"] = "*"
        kwargs["read_only"] = not writable
        super().__init__(*args, **kwargs)

    def __getattribute__(self, name: str) -> Any:
        if (
            name
            not in [
                "validators",
                "get_validators",
                "run_validators",
                "validate_empty_values",
            ]
            or not self.field
        ):
            return super().__getattribute__(name)

        return getattr(self.field, name)

    def bind(self, field_name, parent):
        # The method name defaults to `get_{field_name}`.
        if self.method_name is None:
            self.method_name = "get_{field_name}".format(field_name=field_name)
        super().bind(field_name, parent)
        if self.field is not None:
            set_override(self, "field", self.field)
            # set params for child field
            self.field.parent = self.parent
            self.field.field_name = self.field_name

    def to_representation(self, obj):
        method = getattr(self.parent, self.method_name)
        value = method(obj)
        if not self.field or value is None:
            return value
        else:
            return self.field.to_representation(value)

    def get_default(self):
        return {self.field_name: self.field.get_default}

    def to_internal_value(self, data):
        return {self.field_name: self.field.to_internal_value(data)}

    def run_validation(self, data):
        return {self.field_name: self.field.run_validation(data)}


class Base64ImageField(Base64ExampleMixin, extra_fields.Base64ImageField):
    ...


# TODO - provide example for this one
# class HybridImageField(HybridImageField):
#    ...


class Base64FileField(Base64ExampleMixin, extra_fields.Base64FileField):
    ...


class RangeField(RangeExampleMixin, extra_fields.RangeField):
    ...


@extend_schema_field(IntegerField())
class IntegerRangeField(RangeField):
    child_class = IntegerField
    default_child_attrs = {}
    range_type = NumericRange


@extend_schema_field(FloatField())
class FloatRangeField(RangeField):
    child_class = FloatField
    default_child_attrs = {}
    range_type = NumericRange


@extend_schema_field(DecimalField(**{"max_digits": None, "decimal_places": None}))
class DecimalRangeField(RangeField):
    child_class = DecimalField
    default_child_attrs = {"max_digits": None, "decimal_places": None}
    range_type = NumericRange


@extend_schema_field(DateTimeField())
class DateTimeRangeField(RangeField):
    child_class = DateTimeField
    default_child_attrs = {}
    range_type = DateTimeTZRange


@extend_schema_field(DateField())
class DateRangeField(RangeField):
    child_class = DateField
    default_child_attrs = {}
    range_type = DateRange



from healthcare_api import (
    models,
    serializers,
)
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.permissions import (
    IsAdminUser,
    IsAuthenticated,
)
from rest_framework.response import Response

from django.shortcuts import get_object_or_404

from audoma.decorators import audoma_action
from audoma.drf import mixins
from audoma.drf.viewsets import GenericViewSet


# This viewset shows how audoma works for standard drf's action and custom audoma mixins
# It also has bulk operations included
class PatientViewset(
    mixins.ListModelMixin,
    mixins.BulkCreateModelMixin,
    mixins.BulkUpdateModelMixin,
    mixins.UpdateModelMixin,
    GenericViewSet,
):

    permission_classes = [IsAuthenticated, IsAdminUser]

    serializer_class = serializers.PatientReadSerializer
    common_collect_serializer_class = serializers.PatientWriteSerializer
    get_files_serializer_class = serializers.PatientFilesDetailSerializer
    queryset = models.Patient.objects.all()
    lookup_url_kwarg = "pk"

    def filter_queryset(self, queryset):
        if self.request.method in ["PUT", "PATCH"] and isinstance(
            self.request.data, list
        ):
            ids = [d[self.lookup_url_kwarg] for d in self.request.data]
            queryset = queryset.filter(id__in=ids)
        return queryset

    @action(methods=["GET"], detail=True)
    def get_files(self, request, pk):
        files = get_object_or_404(models.PatientFiles, patient__id=pk)
        serializer = self.get_serializer(instance=files)
        return Response(serializer.data, status=status.HTTP_200_OK)


class SpecializatioNViewSet(mixins.ListModelMixin, GenericViewSet):
    serializer_class = serializers.SpecializationSerializer
    queryset = models.Specialization.objects.all()


class DoctorViewset(
    mixins.ListModelMixin,
    mixins.CreateModelMixin,
    mixins.UpdateModelMixin,
    GenericViewSet,
):

    permission_classes = [IsAuthenticated, IsAdminUser]
    queryset = models.Doctor.objects.all()
    lookup_url_kwarg = "pk"

    list_serializer_class = serializers.DoctorReadSerializer
    create_serializer_class = serializers.DoctorWriteSerializer
    update_serializer_class = serializers.DoctorWriteSerializer
    partial_update_serializer_class = serializers.DoctorWriteSerializer


class PrescriptionViewSet(
    mixins.ListModelMixin, mixins.CreateModelMixin, GenericViewSet
):
    permission_classes = [IsAuthenticated, IsAdminUser]

    list_serializer_class = serializers.PerscriptionReadSerializer
    common_result_serializer_class = serializers.PerscriptionReadSerializer
    common_collect_serializer_class = serializers.PrescrtiptionWriteSerializer
    queryset = models.Prescription.objects.all()
    lookup_url_kwarg = "pk"
    lookup_field = "id"

    @audoma_action(
        detail=True,
        methods=["post"],
        results=serializers.PerscriptionReadSerializer,
        errors=[models.Prescription.DoesNotExist],
        ignore_view_collectors=True,
    )
    def make_prescription_invalid(self, request, pk=None):
        instance = self.get_object()
        instance.is_valid = False
        instance.save()
        return instance, 200



import datetime
from collections import OrderedDict

from drf_example.v2_urls import router
from drf_spectacular.generators import SchemaGenerator
from healthcare_api import models as health_models
from psycopg2._range import DateRange
from rest_framework.test import APITransactionTestCase

from django.contrib.auth.models import User
from django.urls import reverse


class HealthcareAPITestMixin:
    databases = {"healthcare_api", "default"}


class SchemaTestCaseBase(HealthcareAPITestMixin, APITransactionTestCase):
    def setUp(self) -> None:
        super().setUp()
        patterns = router.urls
        generator = SchemaGenerator(patterns=patterns)
        self.schema = generator.get_schema(request=None, public=True)
        self.redoc_schemas = self.schema["components"]["schemas"]


class BasicTestCase(HealthcareAPITestMixin, APITransactionTestCase):
    def setUp(self):
        # create patients
        super().setUp()
        self.user = User.objects.create(
            username="admin", password="passwd", is_staff=True
        )
        self.non_staff_user = User.objects.create(
            username="anybody", password="passwd", is_staff=False
        )


class PatientViewsetTestCase(BasicTestCase):
    def setUp(self):
        # create patients
        super().setUp()
        contact_data = health_models.ContactData.objects.create(
            **{
                "phone_number": "+48 123 456 123",
                "mobile": "+48 908 787 343",
                "country": health_models.COUNTRY_CHOICES.PL,
                "city": "Lasowice",
            }
        )
        for i in range(3):
            health_models.Patient.objects.create(
                name="z" + "z" * i,
                surname="a" + "a" * i,
                contact_data=contact_data,
                height="189.00",
                weight=89.02,
            )
        self.patients = health_models.Patient.objects.all()

    def test_get_list_success(self):
        url = reverse("patient-list")
        self.client.force_authenticate(user=self.user)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["count"], 3)

    def test_get_list_fail_auth(self):
        url = reverse("patient-list")
        response = self.client.get(url)
        self.assertEqual(response.status_code, 401)
        self.assertEqual(
            response.data["errors"]["detail"],
            "Authentication credentials were not provided.",
        )

    def test_get_list_fail_not_admin_user(self):
        url = reverse("patient-list")
        self.client.force_authenticate(user=self.non_staff_user)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.data["errors"]["detail"],
            "You do not have permission to perform this action.",
        )

    def test_create_single_success(self):
        url = reverse("patient-list")
        self.client.force_authenticate(user=self.user)
        request_data = {
            "name": "TestPatient",
            "surname": "Testowy",
            "contact_data": {
                "phone_number": "+48 12 390 98 34",
                "mobile": "+48 12 390 98 34",
                "country": health_models.COUNTRY_CHOICES.PL,
                "city": "Lasowice",
            },
            "weight": 80.0,
            "height": "179.04",
        }
        response = self.client.post(url, request_data, format="json")
        request_data["contact_data"]["country"] = 1
        request_data["contact_data"] = OrderedDict(request_data["contact_data"])
        self.assertEqual(response.status_code, 201)
        self.assertDictEqual(response.data, request_data)

    def test_create_single_fail_invalid_or_missing_data(self):
        url = reverse("patient-list")
        self.client.force_authenticate(user=self.user)
        request_data = {
            "name": "TestPatient",
            "surname": "Testowy",
            "contact_data": {
                "phone_number": "+48 12 390 98 34",
                "mobile": "+48 12 390 98 34",
                "country": health_models.COUNTRY_CHOICES.PL,
                "city": "Lasowice",
            },
            "weight": 80.0,
            "height": "179.04",
        }
        request_data["weight"] = "testtestest"
        response = self.client.post(url, request_data, format="json")
        self.assertEqual(response.status_code, 400)
        self.assertDictEqual(
            response.data, {"errors": {"weight": ["A valid number is required."]}}
        )

    def test_create_single_fail_auth(self):
        url = reverse("patient-list")
        request_data = {
            "name": "TestPatient",
            "surname": "Testowy",
            "contact_data": {
                "phone_number": "+48 12 390 98 34",
                "mobile": "+48 12 390 98 34",
                "country": health_models.COUNTRY_CHOICES.PL,
                "city": "Lasowice",
            },
            "weight": 80.0,
            "height": "179.04",
        }
        response = self.client.post(url, request_data, format="json")
        request_data["contact_data"]["country"] = 1
        request_data["contact_data"] = OrderedDict(request_data["contact_data"])
        self.assertEqual(response.status_code, 401)
        self.assertEqual(
            response.data["errors"]["detail"],
            "Authentication credentials were not provided.",
        )

    def test_create_single_fail_permission(self):
        url = reverse("patient-list")
        self.client.force_authenticate(user=self.non_staff_user)
        request_data = {
            "name": "TestPatient",
            "surname": "Testowy",
            "contact_data": {
                "phone_number": "+48 12 390 98 34",
                "mobile": "+48 12 390 98 34",
                "country": health_models.COUNTRY_CHOICES.PL,
                "city": "Lasowice",
            },
            "weight": 80.0,
            "height": "179.04",
        }
        response = self.client.post(url, request_data, format="json")
        request_data["contact_data"]["country"] = 1
        request_data["contact_data"] = OrderedDict(request_data["contact_data"])
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.data["errors"]["detail"],
            "You do not have permission to perform this action.",
        )

    def test_create_bulk_success(self):
        url = reverse("patient-list")
        self.client.force_authenticate(user=self.user)
        request_data = [
            {
                "name": "TestPatient",
                "surname": "Testowy",
                "contact_data": {
                    "phone_number": "+48 12 390 98 34",
                    "mobile": "+48 12 390 98 34",
                    "country": health_models.COUNTRY_CHOICES.PL,
                    "city": "Lasowice",
                },
                "weight": 80.0,
                "height": "179.04",
            },
            {
                "name": "TestPatient2",
                "surname": "Testowy2",
                "contact_data": {
                    "phone_number": "+48 12 390 98 34",
                    "mobile": "+48 12 390 98 34",
                    "country": health_models.COUNTRY_CHOICES.PL,
                    "city": "Lasowice",
                },
                "weight": 83.0,
                "height": "171.04",
            },
        ]
        response = self.client.post(url, request_data, format="json")
        first_response_data = request_data[0]
        first_response_data["contact_data"]["country"] = 1
        first_response_data["contact_data"] = OrderedDict(
            first_response_data["contact_data"]
        )
        second_response_data = request_data[1]
        second_response_data["contact_data"]["country"] = 1
        second_response_data["contact_data"] = OrderedDict(
            second_response_data["contact_data"]
        )
        self.assertEqual(response.status_code, 201)
        self.assertIsInstance(response.data, list)
        self.assertDictEqual(response.data[0], first_response_data)
        self.assertDictEqual(response.data[1], second_response_data)

    def test_create_bulk_fail_wrong_or_missing_data(self):
        url = reverse("patient-list")
        self.client.force_authenticate(user=self.user)
        request_data = [
            {
                "name": "TestPatient",
                "contact_data": {
                    "phone_number": "+48 12 390 98 34",
                    "mobile": "+48 12 390 98 34",
                    "country": health_models.COUNTRY_CHOICES.PL,
                    "city": "Lasowice",
                },
                "weight": 80.0,
                "height": "179.04",
            },
            {
                "name": "TestPatient2",
                "surname": "Testowy2",
                "contact_data": {
                    "phone_number": "+48 12 390 98 34",
                    "mobile": "+48 12 390 98 34",
                    "country": health_models.COUNTRY_CHOICES.PL,
                    "city": "Lasowice",
                },
                "weight": "ZHD",
                "height": "171.04",
            },
        ]
        response = self.client.post(url, request_data, format="json")
        self.assertEqual(response.status_code, 400)
        self.assertIsInstance(response.data["errors"], list)
        self.assertDictEqual(
            response.data["errors"][0], {"surname": ["This field is required."]}
        )
        self.assertDictEqual(
            response.data["errors"][1], {"weight": ["A valid number is required."]}
        )

    def test_create_bulk_fail_no_auth(self):
        url = reverse("patient-list")
        request_data = [
            {
                "name": "TestPatient",
                "surname": "Testowy",
                "contact_data": {
                    "phone_number": "+48 12 390 98 34",
                    "mobile": "+48 12 390 98 34",
                    "country": health_models.COUNTRY_CHOICES.PL,
                    "city": "Lasowice",
                },
                "weight": 80.0,
                "height": "179.04",
            },
            {
                "name": "TestPatient2",
                "surname": "Testowy2",
                "contact_data": {
                    "phone_number": "+48 12 390 98 34",
                    "mobile": "+48 12 390 98 34",
                    "country": health_models.COUNTRY_CHOICES.PL,
                    "city": "Lasowice",
                },
                "weight": 83.0,
                "height": "171.04",
            },
        ]
        response = self.client.post(url, request_data, format="json")
        self.assertEqual(response.status_code, 401)
        self.assertEqual(
            response.data["errors"]["detail"],
            "Authentication credentials were not provided.",
        )

    def test_create_bulk_fail_permission(self):
        url = reverse("patient-list")
        self.client.force_authenticate(user=self.non_staff_user)
        request_data = [
            {
                "name": "TestPatient",
                "surname": "Testowy",
                "contact_data": {
                    "phone_number": "+48 12 390 98 34",
                    "mobile": "+48 12 390 98 34",
                    "country": health_models.COUNTRY_CHOICES.PL,
                    "city": "Lasowice",
                },
                "weight": 80.0,
                "height": "179.04",
            },
            {
                "name": "TestPatient2",
                "surname": "Testowy2",
                "contact_data": {
                    "phone_number": "+48 12 390 98 34",
                    "mobile": "+48 12 390 98 34",
                    "country": health_models.COUNTRY_CHOICES.PL,
                    "city": "Lasowice",
                },
                "weight": 83.0,
                "height": "171.04",
            },
        ]
        response = self.client.post(url, request_data, format="json")
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.data["errors"]["detail"],
            "You do not have permission to perform this action.",
        )

    def test_partial_update_single_success(self):
        self.client.force_authenticate(user=self.user)
        instance = health_models.Patient.objects.get(surname="aa")
        url = reverse("patient-detail", kwargs={"pk": instance.id})
        response = self.client.patch(url, {"name": "ThisIsPartialUpdate"})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.data,
            {
                "name": "ThisIsPartialUpdate",
                "surname": "aa",
                "contact_data": OrderedDict(
                    [
                        ("phone_number", "+48 12 345 61 23"),
                        ("mobile", "+48 908 787 343"),
                        ("country", 1),
                        ("city", "Lasowice"),
                    ]
                ),
                "weight": 89.02,
                "height": "189.00",
            },
        )

    def test_partial_update_single_fail_wrong_data(self):
        self.client.force_authenticate(user=self.user)
        instance = health_models.Patient.objects.get(surname="aa")
        url = reverse("patient-detail", kwargs={"pk": instance.id})
        response = self.client.patch(url, {"weight": "WRONG_DATA"})
        self.assertEqual(response.status_code, 400)
        self.assertDictEqual(
            response.data["errors"], {"weight": ["A valid number is required."]}
        )

    def test_partial_update_single_fail_auth(self):
        instance = health_models.Patient.objects.get(surname="aa")
        url = reverse("patient-detail", kwargs={"pk": instance.id})
        response = self.client.patch(url, {"name": "CorrectName"})
        self.assertEqual(response.status_code, 401)
        self.assertEqual(
            response.data["errors"]["detail"],
            "Authentication credentials were not provided.",
        )

    def test_partial_update_single_fail_permission(self):
        self.client.force_authenticate(user=self.non_staff_user)
        instance = health_models.Patient.objects.get(surname="aa")
        url = reverse("patient-detail", kwargs={"pk": instance.id})
        response = self.client.patch(url, {"name": "CorrectName"})
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.data["errors"]["detail"],
            "You do not have permission to perform this action.",
        )

    def test_update_single_success(self):
        self.client.force_authenticate(user=self.user)
        instance = health_models.Patient.objects.get(surname="aa")
        url = reverse("patient-detail", kwargs={"pk": instance.id})
        request_data = {
            "name": "TestPatient",
            "surname": "Testowy",
            "contact_data": {
                "phone_number": "+48 12 390 98 34",
                "mobile": "+48 12 390 98 34",
                "country": health_models.COUNTRY_CHOICES.PL,
                "city": "Lasowice",
            },
            "weight": 80.0,
            "height": "179.04",
        }
        response = self.client.put(url, request_data, format="json")
        self.assertEqual(response.status_code, 200)
        request_data["contact_data"]["country"] = 1
        request_data["contact_data"] = OrderedDict(request_data["contact_data"])
        self.assertDictEqual(response.data, request_data)

    def test_update_single_fail_wrong_data(self):
        self.client.force_authenticate(user=self.user)
        instance = health_models.Patient.objects.get(surname="aa")
        url = reverse("patient-detail", kwargs={"pk": instance.id})
        request_data = {
            "name": "TestPatient",
            "surname": "Testowy",
            "contact_data": {
                "phone_number": "+48 12 390 98 34",
                "mobile": "+48 12 390 98 34",
                "country": health_models.COUNTRY_CHOICES.PL,
                "city": "Lasowice",
            },
            "weight": 80.0,
            "height": "179.04",
        }
        request_data["weight"] = "testtestest"
        response = self.client.put(url, request_data, format="json")
        self.assertEqual(response.status_code, 400)
        self.assertDictEqual(
            response.data, {"errors": {"weight": ["A valid number is required."]}}
        )

    def test_update_single_fail_missing_data(self):
        self.client.force_authenticate(user=self.user)
        instance = health_models.Patient.objects.get(surname="aa")
        url = reverse("patient-detail", kwargs={"pk": instance.id})
        request_data = {
            "name": "TestPatient",
            "contact_data": {
                "phone_number": "+48 12 390 98 34",
                "mobile": "+48 12 390 98 34",
                "country": health_models.COUNTRY_CHOICES.PL,
                "city": "Lasowice",
            },
            "weight": 80.0,
            "height": "179.04",
        }
        response = self.client.put(url, request_data, format="json")
        self.assertEqual(response.status_code, 400)
        self.assertDictEqual(
            response.data, {"errors": {"surname": ["This field is required."]}}
        )

    def test_bulk_update_success(self):
        url = reverse("patient-list")
        self.client.force_authenticate(user=self.user)
        data = health_models.Patient.objects.all()[:2]

        request_data = [
            {
                "pk": data[0].id,
                "name": "TestPatient",
                "surname": "Testowy",
                "contact_data": {
                    "phone_number": "+48 12 390 98 34",
                    "mobile": "+48 12 390 98 34",
                    "country": health_models.COUNTRY_CHOICES.PL,
                    "city": "Lasowice",
                },
                "weight": 80.0,
                "height": "179.04",
            },
            {
                "pk": data[1].id,
                "name": "TestPatient2",
                "surname": "Testowy2",
                "contact_data": {
                    "phone_number": "+48 12 390 98 34",
                    "mobile": "+48 12 390 98 34",
                    "country": health_models.COUNTRY_CHOICES.PL,
                    "city": "Lasowice",
                },
                "weight": 83.0,
                "height": "171.04",
            },
        ]
        response = self.client.put(url, request_data, format="json")
        temp = []
        for r in request_data:
            r.pop("pk")
            temp.append(r)
        request_data = temp
        first_response_data = OrderedDict(request_data[0])
        first_response_data["contact_data"]["country"] = 1
        first_response_data["contact_data"] = OrderedDict(
            first_response_data["contact_data"]
        )
        second_response_data = OrderedDict(request_data[1])
        second_response_data["contact_data"]["country"] = 1
        second_response_data["contact_data"] = OrderedDict(
            second_response_data["contact_data"]
        )
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data, list)
        self.assertDictEqual(response.data[0], first_response_data)
        self.assertDictEqual(response.data[1], second_response_data)

    def test_bulk_update_invalid_and_missing_data(self):
        url = reverse("patient-list")
        self.client.force_authenticate(user=self.user)
        data = health_models.Patient.objects.all()[:2]

        request_data = [
            {
                "pk": data[0].id,
                "name": "TestPatient",
                "surname": "Testowy",
                "contact_data": {
                    "phone_number": "+48 12 390 98 34",
                    "mobile": "+48 12 390 98 34",
                    "country": health_models.COUNTRY_CHOICES.PL,
                    "city": "Lasowice",
                },
                "weight": "asdfqwf",
                "height": "179.04",
            },
            {
                "pk": data[1].id,
                "surname": "Testowy2",
                "contact_data": {
                    "phone_number": "+48 12 390 98 34",
                    "mobile": "+48 12 390 98 34",
                    "country": health_models.COUNTRY_CHOICES.PL,
                    "city": "Lasowice",
                },
                "weight": 83.0,
                "height": "171.04",
            },
        ]
        response = self.client.put(url, request_data, format="json")
        self.assertEqual(response.status_code, 400)
        self.assertDictEqual(
            response.data["errors"][1], {"name": ["This field is required."]}
        )
        self.assertDictEqual(
            response.data["errors"][0], {"weight": ["A valid number is required."]}
        )

    def test_bulk_update_fail_no_auth(self):
        url = reverse("patient-list")
        data = health_models.Patient.objects.all()[:2]

        request_data = [
            {
                "pk": data[0].id,
                "name": "TestPatient",
                "surname": "Testowy",
                "contact_data": {
                    "phone_number": "+48 12 390 98 34",
                    "mobile": "+48 12 390 98 34",
                    "country": health_models.COUNTRY_CHOICES.PL,
                    "city": "Lasowice",
                },
                "weight": "asdfqwf",
                "height": "179.04",
            },
            {
                "pk": data[1].id,
                "surname": "Testowy2",
                "contact_data": {
                    "phone_number": "+48 12 390 98 34",
                    "mobile": "+48 12 390 98 34",
                    "country": health_models.COUNTRY_CHOICES.PL,
                    "city": "Lasowice",
                },
                "weight": 83.0,
                "height": "171.04",
            },
        ]
        response = self.client.put(url, request_data, format="json")
        self.assertEqual(response.status_code, 401)
        self.assertEqual(
            response.data["errors"]["detail"],
            "Authentication credentials were not provided.",
        )

    def test_bulk_update_fail_missing_permissions(self):
        url = reverse("patient-list")
        self.client.force_authenticate(user=self.non_staff_user)
        data = health_models.Patient.objects.all()[:2]

        request_data = [
            {
                "pk": data[0].id,
                "name": "TestPatient",
                "surname": "Testowy",
                "contact_data": {
                    "phone_number": "+48 12 390 98 34",
                    "mobile": "+48 12 390 98 34",
                    "country": health_models.COUNTRY_CHOICES.PL,
                    "city": "Lasowice",
                },
                "weight": "asdfqwf",
                "height": "179.04",
            },
            {
                "pk": data[1].id,
                "surname": "Testowy2",
                "contact_data": {
                    "phone_number": "+48 12 390 98 34",
                    "mobile": "+48 12 390 98 34",
                    "country": health_models.COUNTRY_CHOICES.PL,
                    "city": "Lasowice",
                },
                "weight": 83.0,
                "height": "171.04",
            },
        ]
        response = self.client.put(url, request_data, format="json")
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.data["errors"]["detail"],
            "You do not have permission to perform this action.",
        )

    def test_bulk_partial_update_success(self):
        url = reverse("patient-list")
        self.client.force_authenticate(user=self.user)
        data = health_models.Patient.objects.all()[:2]

        expected_response = [
            {
                "name": "TestPatient",
                "surname": d.surname,
                "contact_data": {
                    "phone_number": "+48 12 345 61 23",
                    "mobile": "+48 908 787 343",
                    "country": health_models.COUNTRY_CHOICES.PL,
                    "city": "Lasowice",
                },
                "weight": d.weight,
                "height": str(d.height),
            }
            for d in data
        ]
        response = self.client.patch(
            url,
            [
                {"pk": data[0].id, "name": "TestPatient"},
                {"pk": data[1].id, "name": "TestPatient"},
            ],
            format="json",
        )
        first_response_data = OrderedDict(expected_response[0])
        first_response_data["contact_data"]["country"] = 1
        first_response_data["contact_data"] = OrderedDict(
            first_response_data["contact_data"]
        )
        second_response_data = OrderedDict(expected_response[1])
        second_response_data["contact_data"]["country"] = 1
        second_response_data["contact_data"] = OrderedDict(
            second_response_data["contact_data"]
        )
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.data, list)
        self.assertDictEqual(response.data[0], first_response_data)
        self.assertDictEqual(response.data[1], second_response_data)

    def test_bulk_partial_update_fail_invalid_data(self):
        url = reverse("patient-list")
        self.client.force_authenticate(user=self.user)
        data = health_models.Patient.objects.all()[:2]
        response = self.client.patch(
            url,
            [
                {"pk": data[0].id, "weight": "TestPatient"},
                {"pk": data[1].id, "weight": "TestPatient"},
            ],
            format="json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertDictEqual(
            response.data["errors"][0], {"weight": ["A valid number is required."]}
        )
        self.assertDictEqual(
            response.data["errors"][1], {"weight": ["A valid number is required."]}
        )

    def test_bulk_partial_update_fail_permission(self):
        url = reverse("patient-list")
        self.client.force_authenticate(user=self.non_staff_user)
        data = health_models.Patient.objects.all()[:2]
        response = self.client.patch(
            url,
            [
                {"pk": data[0].id, "weight": "TestPatient"},
                {"pk": data[1].id, "weight": "TestPatient"},
            ],
            format="json",
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.data["errors"]["detail"],
            "You do not have permission to perform this action.",
        )

    def test_bulk_partial_update_fail_no_auht(self):
        url = reverse("patient-list")
        data = health_models.Patient.objects.all()[:2]
        response = self.client.patch(
            url,
            [
                {"pk": data[0].id, "weight": "TestPatient"},
                {"pk": data[1].id, "weight": "TestPatient"},
            ],
            format="json",
        )
        self.assertEqual(response.status_code, 401)
        self.assertEqual(
            response.data["errors"]["detail"],
            "Authentication credentials were not provided.",
        )

    def test_get_files_success(self):
        patient = health_models.Patient.objects.get(id=self.patients[0].id)
        files = health_models.PatientFiles.objects.create(patient=patient)
        entry = health_models.FileEntry.objects.create(
            title="SecretInformation", content="SecretContent", files=files
        )
        url = reverse("patient-get-files", kwargs={"pk": patient.id})
        self.client.force_authenticate(user=self.user)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.maxDiff = None
        expected_response = {
            "patient": OrderedDict(
                {
                    "name": "z",
                    "surname": "a",
                    "contact_data": OrderedDict(
                        {
                            "phone_number": "+48 12 345 61 23",
                            "mobile": "+48 908 787 343",
                            "country": health_models.COUNTRY_CHOICES.PL,
                            "city": "Lasowice",
                        }
                    ),
                    "weight": 89.02,
                    "height": "189.00",
                }
            ),
            "entries": [
                OrderedDict(
                    {
                        "created_at": entry.created_at.strftime(
                            "%Y-%m-%dT%H:%M:%S.%fZ"
                        ),
                        "modifed_at": entry.modifed_at.strftime(
                            "%Y-%m-%dT%H:%M:%S.%fZ"
                        ),
                        "title": "SecretInformation",
                        "content": "SecretContent",
                    }
                )
            ],
        }
        self.assertDictEqual(response.data, expected_response)

    def test_get_files_fail_not_existing_patient(self):
        url = reverse("patient-get-files", kwargs={"pk": 2137})
        self.client.force_authenticate(user=self.user)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 404)

    def test_get_files_fail_permission(self):
        patient = health_models.Patient.objects.filter(name="z").first()
        health_models.PatientFiles.objects.create(patient=patient)
        url = reverse("patient-get-files", kwargs={"pk": 1})
        self.client.force_authenticate(user=self.non_staff_user)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.data["errors"]["detail"],
            "You do not have permission to perform this action.",
        )

    def test_get_files_fail_no_auth(self):
        url = reverse("patient-get-files", kwargs={"pk": 1})
        response = self.client.get(url)
        self.assertEqual(response.status_code, 401)
        self.assertEqual(
            response.data["errors"]["detail"],
            "Authentication credentials were not provided.",
        )


class PatientViewsetSchemaTestCase(SchemaTestCaseBase):
    def test_get_list_path_schema(self):
        path_schema = self.schema["paths"]["/patient/"]["get"]
        self.assertEqual(
            path_schema["description"],
            "\n\n**Permissions:**\n`IsAuthenticated` & `IsAdminUser`\n+ `IsAuthenticated`: "
            + "*Allows access only to authenticated users.*\n+ `IsAdminUser`: "
            + "*Allows access only to admin users.*",
        )
        self.assertDictEqual(
            path_schema["responses"],
            {
                "200": {
                    "content": {
                        "application/json": {
                            "schema": {
                                "$ref": "#/components/schemas/PaginatedPatientReadList"
                            }
                        }
                    },
                    "description": "",
                }
            },
        )

    def test_get_list_response_schema(self):
        props = self.redoc_schemas["PatientRead"]["properties"]
        self.assertEqual(props["height"]["pattern"], "^\\d{0,4}(?:\\.\\d{0,2})?$")
        self.assertDictEqual(
            props["contact_data"], {"$ref": "#/components/schemas/ContactData"}
        )

    def test_contact_data_response_schema(self):
        props = self.redoc_schemas["ContactData"]["properties"]
        self.assertEqual(props["country"]["enum"], [1, 2])
        self.assertDictEqual(
            props["country"]["x-choices"], {"choices": {1: "Poland", 2: "Germany"}}
        )
        self.assertEqual(
            props["country"]["description"],
            "Filter by country \n * `1` - Poland\n * `2` - Germany\n",
        )
        self.assertEqual(props["phone_number"]["type"], "string")
        self.assertEqual(props["phone_number"]["format"], "tel")
        self.assertEqual(props["mobile"]["type"], "string")
        self.assertEqual(props["mobile"]["format"], "tel")


class DoctorViesetTestCase(BasicTestCase):
    def setUp(self):
        super().setUp()
        health_models.Specialization.objects.create(name="Pediatry")
        health_models.Specialization.objects.create(name="Radiology")
        health_models.Specialization.objects.create(name="Anesthesiology")
        self.specializations = health_models.Specialization.objects.all()
        contact_data = health_models.ContactData.objects.create(
            **{
                "phone_number": "+48 123 456 123",
                "mobile": "+48 908 787 343",
                "country": health_models.COUNTRY_CHOICES.PL,
                "city": "Lasowice",
            }
        )
        for i in range(3):
            salary = 3000.0 * i + 1000.0 * i
            doc = health_models.Doctor.objects.create(
                name="z" + "z" * i,
                surname="a" + "a" * i,
                contact_data=contact_data,
                salary=salary,
            )
            doc.specialization.add(self.specializations[i])
            doc.save()
        self.doctors = health_models.Doctor.objects.all()

    def test_get_list_success(self):
        url = reverse("doctor-list")
        self.client.force_authenticate(user=self.user)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["count"], 3)
        results = response.data["results"]

        self.assertDictEqual(
            results[0]["specialization"][0],
            OrderedDict({"id": self.specializations[0].id, "name": "Pediatry"}),
        )
        self.assertDictEqual(
            results[1]["specialization"][0],
            OrderedDict({"id": self.specializations[1].id, "name": "Radiology"}),
        )
        self.assertDictEqual(
            results[2]["specialization"][0],
            OrderedDict({"id": self.specializations[2].id, "name": "Anesthesiology"}),
        )

    def test_get_list_no_auth(self):
        url = reverse("doctor-list")
        response = self.client.get(url)
        self.assertEqual(response.status_code, 401)
        self.assertEqual(
            response.data["errors"]["detail"],
            "Authentication credentials were not provided.",
        )

    def test_get_list_lack_permissions(self):
        url = reverse("doctor-list")
        self.client.force_authenticate(user=self.non_staff_user)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.data["errors"]["detail"],
            "You do not have permission to perform this action.",
        )

    def test_create_doctor_success(self):
        url = reverse("doctor-list")
        self.client.force_authenticate(user=self.user)
        data = {
            "name": "TeestDoctor",
            "surname": "Creepy",
            "contact_data": {
                "phone_number": "+48 455 123 432",
                "mobile": "+48 455 123 432",
                "country": health_models.COUNTRY_CHOICES.PL,
                "city": "Gliwice",
            },
            "salary": "5000.00",
            "specialization": [s.id for s in self.specializations],
        }
        response = self.client.post(url, data, format="json")
        self.assertEqual(response.status_code, 201)
        self.assertEqual(
            response.data["specialization"], [s.id for s in self.specializations]
        )

    def test_create_doctor_failure_missing_data(self):
        url = reverse("doctor-list")
        self.client.force_authenticate(user=self.user)
        data = {
            "name": "TeestDoctor",
            "surname": "Creepy",
            "contact_data": {
                "phone_number": "+48 455 123 432",
                "mobile": "+48 455 123 432",
                "country": health_models.COUNTRY_CHOICES.PL,
                "city": "Gliwice",
            },
            "specialization": [1, 2],
        }
        response = self.client.post(url, data, format="json")
        self.assertEqual(response.status_code, 400)
        self.assertDictEqual(
            response.data, {"errors": {"salary": ["This field is required."]}}
        )

    def test_create_doctor_failure_wrong_data(self):
        url = reverse("doctor-list")
        self.client.force_authenticate(user=self.user)
        data = {
            "name": "TeestDoctor",
            "surname": "Creepy",
            "contact_data": {
                "phone_number": "+48 455 123 432",
                "mobile": "+48 455 123 432",
                "country": health_models.COUNTRY_CHOICES.PL,
                "city": "Gliwice",
            },
            "salary": "TEST",
            "specialization": [1, 2],
        }
        response = self.client.post(url, data, format="json")
        self.assertEqual(response.status_code, 400)
        self.assertDictEqual(
            response.data, {"errors": {"salary": ["A valid number is required."]}}
        )

    def test_create_doctor_failure_no_auth(self):
        url = reverse("doctor-list")
        data = {
            "name": "TeestDoctor",
            "surname": "Creepy",
            "contact_data": {
                "phone_number": "+48 455 123 432",
                "mobile": "+48 455 123 432",
                "country": health_models.COUNTRY_CHOICES.PL,
                "city": "Gliwice",
            },
            "specialization": [1, 2],
            "salary": "5000.00",
        }
        response = self.client.post(url, data, format="json")
        self.assertEqual(response.status_code, 401)
        self.assertEqual(
            response.data["errors"]["detail"],
            "Authentication credentials were not provided.",
        )

    def test_create_doctor_failure_missing_permissions(self):
        url = reverse("doctor-list")
        self.client.force_authenticate(user=self.non_staff_user)
        data = {
            "name": "TeestDoctor",
            "surname": "Creepy",
            "contact_data": {
                "phone_number": "+48 455 123 432",
                "mobile": "+48 455 123 432",
                "country": health_models.COUNTRY_CHOICES.PL,
                "city": "Gliwice",
            },
            "salary": "5000.00",
            "specialization": [1, 2],
        }
        response = self.client.post(url, data, format="json")
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.data["errors"]["detail"],
            "You do not have permission to perform this action.",
        )

    def test_update_doctor_success(self):
        url = reverse("doctor-detail", kwargs={"pk": self.doctors[0].id})
        self.client.force_authenticate(user=self.user)
        data = {
            "name": "TeestDoctor",
            "surname": "Creepy",
            "contact_data": {
                "phone_number": "+48 455 123 432",
                "mobile": "+48 455 123 432",
                "country": health_models.COUNTRY_CHOICES.PL,
                "city": "Gliwice",
            },
            "salary": "5000.00",
            "specialization": [s.id for s in self.doctors[0].specialization.all()],
        }
        response = self.client.put(url, data, format="json")
        self.assertEqual(response.status_code, 200)
        data["contact_data"] = OrderedDict(data["contact_data"])
        self.assertDictEqual(response.data, data)

    def test_update_doctor_fail_missing_data(self):
        url = reverse("doctor-detail", kwargs={"pk": self.doctors[0].id})
        self.client.force_authenticate(user=self.user)
        data = {
            "name": "TeestDoctor",
            "surname": "Creepy",
            "contact_data": {
                "phone_number": "+48 455 123 432",
                "mobile": "+48 455 123 432",
                "country": health_models.COUNTRY_CHOICES.PL,
                "city": "Gliwice",
            },
            "specialization": [s.id for s in self.doctors[0].specialization.all()],
        }
        response = self.client.put(url, data, format="json")
        self.assertEqual(response.status_code, 400)
        self.assertDictEqual(
            response.data, {"errors": {"salary": ["This field is required."]}}
        )

    def test_update_doctor_fail_wrong_data(self):
        url = reverse("doctor-detail", kwargs={"pk": self.doctors[0].id})
        self.client.force_authenticate(user=self.user)
        data = {
            "name": "TeestDoctor",
            "surname": "Creepy",
            "contact_data": {
                "phone_number": "+48 455 123 432",
                "mobile": "+48 455 123 432",
                "country": health_models.COUNTRY_CHOICES.PL,
                "city": "Gliwice",
            },
            "salary": "TEST",
            "specialization": [s.id for s in self.doctors[0].specialization.all()],
        }
        response = self.client.put(url, data, format="json")
        self.assertEqual(response.status_code, 400)
        self.assertDictEqual(
            response.data, {"errors": {"salary": ["A valid number is required."]}}
        )

    def test_partial_update_doctor_success(self):
        url = reverse("doctor-detail", kwargs={"pk": self.doctors[0].pk})
        self.client.force_authenticate(user=self.user)
        data = {
            "name": "TeestDoctor",
            "surname": "Creepy",
            "contact_data": {
                "phone_number": "+48 12 345 61 23",
                "mobile": "+48 908 787 343",
                "country": health_models.COUNTRY_CHOICES.PL,
                "city": "Lasowice",
            },
            "salary": "0.00",
            "specialization": [s.id for s in self.doctors[0].specialization.all()],
        }
        response = self.client.patch(
            url,
            {
                "name": "TeestDoctor",
                "surname": "Creepy",
            },
            format="json",
        )
        self.assertEqual(response.status_code, 200)
        data["contact_data"] = OrderedDict(data["contact_data"])
        self.assertDictEqual(response.data, data)

    def test_partial_update_doctor_failure_wrong_data(self):
        url = reverse("doctor-detail", kwargs={"pk": self.doctors[0].pk})
        self.client.force_authenticate(user=self.user)
        response = self.client.patch(
            url,
            {
                "salary": "Z@#!@FASD",
            },
            format="json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertDictEqual(
            response.data, {"errors": {"salary": ["A valid number is required."]}}
        )

    def test_partial_update_fail_no_auth(self):
        url = reverse("doctor-detail", kwargs={"pk": 1})
        response = self.client.patch(
            url,
            {
                "salary": "5000.0",
            },
            format="json",
        )
        self.assertEqual(response.status_code, 401)
        self.assertEqual(
            response.data["errors"]["detail"],
            "Authentication credentials were not provided.",
        )

    def test_partial_update_fail_no_permissions(self):
        url = reverse("doctor-detail", kwargs={"pk": 1})
        self.client.force_authenticate(user=self.non_staff_user)
        response = self.client.patch(
            url,
            {
                "salary": "5000.0",
            },
            format="json",
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.data["errors"]["detail"],
            "You do not have permission to perform this action.",
        )


class DoctorSchemaTestCase(SchemaTestCaseBase):
    def test_x_choices_link_generated_properly(self):
        enpoint_schema = self.redoc_schemas["DoctorWrite"]
        x_choices_link = enpoint_schema["properties"]["specialization"]["x-choices"]
        self.assertDictEqual(
            x_choices_link,
            {
                "operationRef": "#/paths/~1specialization~1",
                "value": "$response.body#results/*/id",
                "display": "$response.body#results/*/name",
            },
        )

    def test_responses_for_dictor_viewset_created_success(self):
        path_schema = self.schema["paths"]["/doctor/"]
        get_method_schema = path_schema["get"]
        self.assertEqual(get_method_schema["operationId"], "doctor_list")
        self.assertDictEqual(
            get_method_schema["responses"],
            {
                "200": {
                    "content": {
                        "application/json": {
                            "schema": {
                                "$ref": "#/components/schemas/PaginatedDoctorReadList"
                            }
                        }
                    },
                    "description": "",
                }
            },
        )


class PrescriptionViewsetTestCase(BasicTestCase):
    def setUp(self):
        super().setUp()
        # create_doctor
        self.doctor_data = {
            "name": "John",
            "surname": "Kowalski",
            "contact_data": health_models.ContactData.objects.create(
                **{
                    "phone_number": "+48 455 123 432",
                    "mobile": "+48 455 123 432",
                    "country": health_models.COUNTRY_CHOICES.PL,
                    "city": "Gliwice",
                }
            ),
            "salary": "5000.0",
        }
        self.patient_data = {
            "name": "John",
            "surname": "Kowalski",
            "contact_data": health_models.ContactData.objects.create(
                **{
                    "phone_number": "+48 123 456 123",
                    "mobile": "+48 908 787 343",
                    "country": health_models.COUNTRY_CHOICES.PL,
                    "city": "Lasowice",
                }
            ),
            "weight": 80.0,
            "height": "189.00",
        }
        self.doctor = health_models.Doctor.objects.create(**self.doctor_data)
        self.doctor.specialization.add(
            health_models.Specialization.objects.create(name="pediatry")
        )
        self.patient = health_models.Patient.objects.create(**self.patient_data)
        self.prescription = health_models.Prescription.objects.create(
            issued_by=self.doctor,
            issued_for=self.patient,
            drugs=[{"testDrug1": 2}, {"TestDrug2": 3}],
            usable_in=DateRange(
                lower=datetime.date.today(),
                upper=datetime.date.today() + datetime.timedelta(days=3),
            ),
            issued_in="Some Test Hospital Somewhere",
        )

    def test_get_list_success(self):
        self.client.force_authenticate(self.user)
        url = reverse("prescription-list")
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        data = response.data["results"]
        self.assertEqual(response.data["count"], 1)
        self.assertDictEqual(
            data[0],
            OrderedDict(
                {
                    "issued_by": self.doctor.id,
                    "issued_for": self.patient.id,
                    "drugs": [{"testDrug1": "2"}, {"TestDrug2": "3"}],
                    "usable_in": {
                        "lower": str(self.prescription.usable_in.lower),
                        "upper": str(self.prescription.usable_in.upper),
                        "bounds": "[)",
                    },
                    "issued_in": "Some Test Hospital Somewhere",
                    "is_valid": True,
                }
            ),
        )

    def test_get_list_no_auth(self):
        url = reverse("prescription-list")
        response = self.client.get(url)
        self.assertEqual(response.status_code, 401)
        self.assertEqual(
            response.data["errors"]["detail"],
            "Authentication credentials were not provided.",
        )

    def test_get_list_permission_denied(self):
        self.client.force_authenticate(self.non_staff_user)
        url = reverse("prescription-list")
        response = self.client.get(url)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.data["errors"]["detail"],
            "You do not have permission to perform this action.",
        )

    def test_create_prescription_success(self):
        self.client.force_authenticate(self.user)
        url = reverse("prescription-list")
        response = self.client.post(
            url,
            data={
                "issued_by": self.doctor.id,
                "issued_for": self.patient.id,
                "drugs": [{"GreatDrug": "12"}],
                "usable_in": {
                    "lower": str(self.prescription.usable_in.lower),
                    "upper": str(self.prescription.usable_in.upper),
                    "bounds": "[)",
                },
                "issued_in": "Some Test Hospital Somewhere",
            },
            format="json",
        )
        self.assertEqual(response.status_code, 201)
        self.assertDictEqual(
            response.data,
            {
                "issued_by": self.doctor.id,
                "issued_for": self.patient.id,
                "drugs": [{"GreatDrug": "12"}],
                "usable_in": {
                    "lower": str(self.prescription.usable_in.lower),
                    "upper": str(self.prescription.usable_in.upper),
                    "bounds": "[)",
                },
                "issued_in": "Some Test Hospital Somewhere",
                "is_valid": True,
            },
        )

    def test_create_prescription_fail_no_auth(self):
        url = reverse("prescription-list")
        response = self.client.post(
            url,
            data={
                "issued_by": self.doctor.id,
                "issued_for": self.patient.id,
                "drugs": [{"GreatDrug": "12"}],
                "usable_in": {
                    "lower": str(self.prescription.usable_in.lower),
                    "upper": str(self.prescription.usable_in.upper),
                    "bounds": "[)",
                },
                "issued_in": "Some Test Hospital Somewhere",
            },
            format="json",
        )
        self.assertEqual(response.status_code, 401)
        self.assertEqual(
            response.data["errors"]["detail"],
            "Authentication credentials were not provided.",
        )

    def test_create_prescription_fail_no_permissions(self):
        self.client.force_authenticate(self.non_staff_user)
        url = reverse("prescription-list")
        response = self.client.post(
            url,
            data={
                "issued_by": self.doctor.id,
                "issued_for": self.patient.id,
                "drugs": [{"GreatDrug": "12"}],
                "usable_in": {
                    "lower": str(self.prescription.usable_in.lower),
                    "upper": str(self.prescription.usable_in.upper),
                    "bounds": "[)",
                },
                "issued_in": "Some Test Hospital Somewhere",
            },
            format="json",
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.data["errors"]["detail"],
            "You do not have permission to perform this action.",
        )

    def test_make_prescription_invalid_success(self):
        self.client.force_authenticate(self.user)
        url = reverse(
            "prescription-make-prescription-invalid",
            kwargs={"pk": self.prescription.id},
        )
        response = self.client.post(url)
        self.assertEqual(response.status_code, 200)
        self.assertDictEqual(
            response.data,
            {
                "issued_by": self.doctor.id,
                "issued_for": self.patient.id,
                "drugs": [{"testDrug1": "2"}, {"TestDrug2": "3"}],
                "usable_in": {
                    "lower": str(self.prescription.usable_in.lower),
                    "upper": str(self.prescription.usable_in.upper),
                    "bounds": "[)",
                },
                "issued_in": "Some Test Hospital Somewhere",
                "is_valid": False,
            },
        )



# Generated by Django 4.1.5 on 2023-02-17 13:46
from decimal import Decimal

import django.core.validators
from django.db import migrations

import audoma.django.db.fields


class Migration(migrations.Migration):

    dependencies = [
        ("audoma_api", "0006_cartag"),
    ]

    operations = [
        migrations.AlterField(
            model_name="examplemodel",
            name="decimal",
            field=audoma.django.db.fields.DecimalField(
                decimal_places=2,
                max_digits=10,
                validators=[django.core.validators.MinValueValidator(Decimal("1.0"))],
            ),
        ),
    ]



"""
Django settings for drf_example project.

Generated by 'django-admin startproject' using Django 4.0.2.

For more information on this file, see
https://docs.djangoproject.com/en/4.0/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/4.0/ref/settings/
"""
import os
from pathlib import Path


# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/4.0/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = os.environ.get(
    "SECRET_KEY", "django-insecure-+37#%@+#3j18rkpq-w9u&s5s5hl$vt#+4we2697_7aupm8n96p"
)
# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = ["*"]


# Application definition

INSTALLED_APPS = [
    "audoma_api",
    "healthcare_api",
    "djmoney",
    "drf_spectacular",
    "rest_framework",
    "django_filters",
    "django_extensions",
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.postgres",
    "django_nose",
    "rest_framework.authtoken",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "drf_example.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "drf_example.wsgi.application"


# Database
# https://docs.djangoproject.com/en/4.0/ref/settings/#databases


# Password validation
# https://docs.djangoproject.com/en/4.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]


# Internationalization
# https://docs.djangoproject.com/en/4.0/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/4.0/howto/static-files/

STATIC_URL = "static/"

# Default primary key field type
# https://docs.djangoproject.com/en/4.0/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

REST_FRAMEWORK = {
    # YOUR SETTINGS
    "DEFAULT_SCHEMA_CLASS": "audoma.openapi.AudomaAutoSchema",
    "DEFAULT_FILTER_BACKENDS": ("django_filters.rest_framework.DjangoFilterBackend",),
    "DEFAULT_RENDERER_CLASSES": [
        "rest_framework.renderers.JSONRenderer",
    ],
    "DEFAULT_AUTHENTICATION_CLASSES": [
        "rest_framework.authentication.TokenAuthentication",
    ],
}

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "db.sqlite3",
        "TEST_NAME": "test_db.sqlite3",
    },
    "audoma_api": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "db.sqlite3",
        "TEST_NAME": "test_db.sqlite3",
    },
    "healthcare_api": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": os.environ.get("PSQL_DB_NAME", "audoma_healthcare_example"),
        "USER": os.environ.get("PSQL_USER", "postgres"),
        "PASSWORD": os.environ.get("PSQL_PASSWORD", "password"),
        "HOST": os.environ.get("PSQL_HOST", "postgres"),
        "PORT": os.environ.get("PSQL_PORT", 5432),
        "TEST_NAME": "audoma_healthcare_example_tests",
    },
}

DATABASE_ROUTERS = [
    "routers.db_routers.HealthCareApiRouter",
    "routers.db_routers.AudomaApiRouter",
]

# Use it if you want to create schema only based on paths that starts with specific keyword
# and then add "PREPROCESSING_HOOKS" to SPECTACULAR_SETTINGS as commented below
# SCHEMA_PATTERN_PREFIX = 'api'

SPECTACULAR_SETTINGS = {
    "TITLE": "audoma API",
    "DESCRIPTION": "API Automatic Documentation Maker - YASG/SPECTACULAR wrapper",
    "VERSION": "1.0.0",
    # 'PREPROCESSING_HOOKS': ['audoma.hooks.preprocess_include_path_format'],
    # OTHER SETTINGS
    "COMPONENT_SPLIT_REQUEST": True,
}

CURRENCIES = ("USD", "EUR", "GBP", "JPY", "CNY", "INR", "AUD", "NZD", "CHF")
PHONENUMBER_DEFAULT_FORMAT = "INTERNATIONAL"

TEST_RUNNER = "django_nose.NoseTestSuiteRunner"



from rest_framework import routers
from rest_framework.routers import *  # noqa: F403, F401


class SimpleRouter(routers.SimpleRouter):
    routes = [
        # List route.
        routers.Route(
            url=r"^{prefix}{trailing_slash}$",
            mapping={
                "get": "list",
                "post": "create",
                "put": "bulk_update",
                "patch": "partial_bulk_update",
                "delete": "bulk_destroy",
            },
            name="{basename}-list",
            detail=False,
            initkwargs={"suffix": "List"},
        ),
        # Dynamically generated list routes. Generated using
        # @action(detail=False) decorator on methods of the viewset.
        routers.DynamicRoute(
            url=r"^{prefix}/{url_path}{trailing_slash}$",
            name="{basename}-{url_name}",
            detail=False,
            initkwargs={},
        ),
        # Detail route.
        routers.Route(
            url=r"^{prefix}/{lookup}{trailing_slash}$",
            mapping={
                "get": "retrieve",
                "put": "update",
                "patch": "partial_update",
                "delete": "destroy",
            },
            name="{basename}-detail",
            detail=True,
            initkwargs={"suffix": "Instance"},
        ),
        # Dynamically generated detail routes. Generated using
        # @action(detail=True) decorator on methods of the viewset.
        routers.DynamicRoute(
            url=r"^{prefix}/{lookup}/{url_path}{trailing_slash}$",
            name="{basename}-{url_name}",
            detail=True,
            initkwargs={},
        ),
    ]


class DefaultRouter(SimpleRouter, routers.DefaultRouter):
    ...


# DEPRECATED - to be deleted
class BulkRouter(DefaultRouter):
    """
    Map http methods to actions defined on the bulk mixins.
    """

    # routes = copy.deepcopy(SimpleRouter.routes)
    # routes[0].mapping.update(
    #     {
    #         "put": "bulk_update",
    #         "patch": "partial_bulk_update",
    #         "delete": "bulk_destroy",
    #     }
    # )



import datetime
import random
from decimal import Decimal
from typing import (
    Any,
    Type,
)

import exrex

from django.core import validators


class DEFAULT:
    pass


class Example:
    """
    Class that represents an example for a field.
    It allows to add example to the field during initialization.
    """

    def __init__(self, field, example=DEFAULT) -> None:
        self.field = field
        self.example = example

    def generate_value(self) -> Type[DEFAULT]:
        return DEFAULT

    def get_value(self) -> Any:
        if self.example is not DEFAULT:
            if callable(self.example):
                return self.example()
            return self.example
        return self.generate_value()

    def to_representation(self, value) -> Any:
        return self.field.to_representation(value)


class NumericExample(Example):
    def generate_value(self) -> float:
        """
        Extracts information from the field and generates a random value
        based on min_value and max_value.

        Returns:
            Random value between min_value and max_value
        """
        min_val = float(getattr(self.field, "min_value", 1) or 1)
        max_val = float(getattr(self.field, "max_value", 1000) or 1000)
        decimal_places = getattr(self.field, "decimal_places", None)
        if decimal_places:
            fmt = f".{decimal_places}f"
            return Decimal(f"{(max_val):{fmt}}")
        ret = random.uniform(min_val, max_val)
        return ret


class RegexExample(Example):
    def generate_value(self) -> str:
        """
        Extracts information from the field and generates a random value
        based on field's  RegexValidators.

        Returns:
            Generated regex string if regex is found, otherwise returns None
        """
        regex_validators = [
            validator
            for validator in self.field.validators
            if isinstance(validator, validators.RegexValidator)
        ]
        if regex_validators:
            regex_validator = regex_validators[0]
            return exrex.getone(regex_validator.regex.pattern)
        return None


class _DateRelatedExampleMixin:
    def generate_value(self) -> datetime.datetime:
        return datetime.datetime.now() - datetime.timedelta(
            days=random.randint(0, 20),
            hours=random.randint(0, 10),
            minutes=random.randint(0, 40),
        )


class DateExample(_DateRelatedExampleMixin, Example):
    def generate_value(self) -> datetime.date:
        return super().generate_value().date()


class TimeExample(_DateRelatedExampleMixin, Example):
    def generate_value(self) -> datetime.date:
        return super().generate_value().time()


class DateTimeExample(_DateRelatedExampleMixin, Example):
    ...


class Base64Example(Example):
    def generate_value(self) -> float:
        """
        Extracts information from the field and generates a random value
        based on min_value and max_value.

        Returns:
            Random value between min_value and max_value
        """
        max_length = float(getattr(self.field, "max_length", 1) or 1)
        length = max_length
        return "%030x" % random.randrange(16**length)


class RangeExample(Example):
    def generate_value(self) -> float:
        """
        Extracts information from the field and generates a random value
        based on min_value and max_value.

        Returns:
            Random value between min_value and max_value
        """
        child_field = getattr(self.field, "child", None)
        example_class = (
            child_field.audoma_example_class
            if child_field is not None
            else NumericExample
        )
        lower, upper = (
            example_class(field=child_field).generate_value(),
            example_class(field=child_field).generate_value(),
        )
        lower, upper = (upper, lower) if lower > upper else (lower, upper)
        return {"lower": lower, "upper": upper}



# Generated by Django 4.1.3 on 2022-12-01 09:49

import django.db.models.deletion
from django.db import (
    migrations,
    models,
)

import audoma.django.db.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="ContactData",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "phone_number",
                    audoma.django.db.fields.PhoneNumberField(
                        max_length=128, region="PL"
                    ),
                ),
                (
                    "mobile",
                    audoma.django.db.fields.PhoneNumberField(
                        max_length=128, region="PL"
                    ),
                ),
                (
                    "country",
                    audoma.django.db.fields.CharField(
                        choices=[(1, "Poland"), (2, "Germany")], max_length=100
                    ),
                ),
                ("city", audoma.django.db.fields.CharField(max_length=200)),
            ],
        ),
        migrations.CreateModel(
            name="Patient",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", audoma.django.db.fields.CharField(max_length=255)),
                ("surname", audoma.django.db.fields.CharField(max_length=255)),
                ("weight", audoma.django.db.fields.FloatField()),
                (
                    "height",
                    audoma.django.db.fields.DecimalField(
                        decimal_places=2, max_digits=6
                    ),
                ),
                (
                    "contact_data",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="healthcare_api.contactdata",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Specialization",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", audoma.django.db.fields.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name="PatientFiles",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "patient",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="healthcare_api.patient",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="FileEntry",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    audoma.django.db.fields.DateTimeField(auto_now_add=True),
                ),
                ("modifed_at", audoma.django.db.fields.DateTimeField(auto_now=True)),
                ("title", audoma.django.db.fields.CharField(max_length=200)),
                ("content", audoma.django.db.fields.TextField()),
                (
                    "files",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="healthcare_api.patientfiles",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Doctor",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", audoma.django.db.fields.CharField(max_length=255)),
                ("surname", audoma.django.db.fields.CharField(max_length=255)),
                (
                    "salary_currency",
                    audoma.django.db.fields.CurrencyField(
                        choices=[
                            ("AUD", "Australian Dollar"),
                            ("GBP", "British Pound"),
                            ("CNY", "Chinese Yuan"),
                            ("EUR", "Euro"),
                            ("INR", "Indian Rupee"),
                            ("JPY", "Japanese Yen"),
                            ("NZD", "New Zealand Dollar"),
                            ("CHF", "Swiss Franc"),
                            ("USD", "US Dollar"),
                        ],
                        default="PLN",
                        max_length=3,
                    ),
                ),
                (
                    "salary",
                    audoma.django.db.fields.MoneyField(
                        decimal_places=2, default_currency="PLN", max_digits=10
                    ),
                ),
                (
                    "contact_data",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="healthcare_api.contactdata",
                    ),
                ),
                (
                    "specialization",
                    models.ManyToManyField(to="healthcare_api.specialization"),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]
